module VariationalAssimilation

using PythonCall

function __init__()
	global sp = pyimport("sympy")
end

macro sympy_register(func)
    if isdefined(Base, func) && hasproperty(sp, func)
        quote
            import Base.$(func)
            Base.$(func)(x::Py) = sp.$(func)(x)
        end
    else
        throw(ArgumentError("Function is undefined in either Base or SymPy."))
    end
end

include("./varassim/varassim_structs.jl")
include("./varassim/varassim_nlp.jl")
include("./varassim/gen_ipopt_callbacks.jl")
include("./discretizations/discretizations.jl")
include("./dpse/dpse.jl")

export @sympy_register
export callback_module_str

export VarAssimVectorField, VarAssimTimestepResidualCost
export VarAssimObservations, VarAssimBounds
export VarAssimProblem

export timestep_residual, num_intermediate, num_constraints
export SimpsonHermite, HLGL5

export DPSEProblemMeta, build_dpse_problem

end # module VariationalAssimilation
