abstract type VarAssimDiscretization end;

struct Trapezoid <: VarAssimDiscretization end;
struct SimpsonHermite <: VarAssimDiscretization end;
struct HLGL5 <: VarAssimDiscretization end;

num_intermediate(disc::Trapezoid) = 0
num_intermediate(disc::SimpsonHermite) = 1
num_intermediate(disc::HLGL5) = 3

num_constraints(disc::Trapezoid, D) = D
num_constraints(disc::SimpsonHermite, D) = 2*D
num_constraints(disc::HLGL5, D) = 4*D

#simple collocation_constraint which is linear in states and derivatives
function collocation_constraint(ucoeff, Fucoeff, dt, ublock, Fublock)
    D = length(ublock[1])
    R = length(ucoeff)
    sum(ucoeff[j] .* ublock[j] .+ repeat([dt], D) .* (Fucoeff[j] .* Fublock[j]) for j in 1:R)
end

function timestep_residual(disc::SimpsonHermite, vavf::VarAssimVectorField, ublock, ocblock, p, scblock, dt)
    R = num_intermediate(disc)
    Fublock = [vavf.vf(ublock[r], ocblock[r], p, scblock[r]) for r=1:(R+2)] 

    simp_ucoeff = [1.0, 0.0, -1.0]
    simp_Fucoeff = [1.0/6, 4/6, 1/6]
    herm_ucoeff = [0.5, -1.0, 0.5]
    herm_Fucoeff = [1/8, 0.0, -1/8]

    simp = collocation_constraint(simp_ucoeff, simp_Fucoeff, dt, ublock, Fublock)
    herm = collocation_constraint(herm_ucoeff, herm_Fucoeff, dt, ublock, Fublock)

    vcat(simp, herm)
end

function timestep_residual(disc::HLGL5, vavf::VarAssimVectorField, ublock, ocblock, p, scblock, dt)
    R = num_intermediate(disc)
    Fublock = [vavf.vf(ublock[r], ocblock[r], p, scblock[r]) for r=1:(R+2)] 

    x2_ucoeff = [(39*sqrt(21)+231)/686, -1.0, (224/686), 0.0, (-39*sqrt(21)+231)/686]
    x2_Fucoeff = [(3*sqrt(21)+21)/686, 0.0, -16*sqrt(21)/686, 0.0, (3*sqrt(21) - 21)/686]
    x4_ucoeff = [(-39*sqrt(21)+231)/686, 0.0, (224/686), -1.0, (39*sqrt(21)+231)/686]
    x4_Fucoeff = [(-3*sqrt(21)+21)/686, 0.0, 16*sqrt(21)/686, 0.0, (-3*sqrt(21) - 21)/686]

    ζ2_ucoeff = [(32*sqrt(21)+180)/360, 0.0, -64*sqrt(21)/360, 0.0, (32*sqrt(21) - 180)/360]
    ζ2_Fucoeff = [(9+sqrt(21))/360, 98/360, 64/360, 0.0, (9-sqrt(21))/360]
    ζ4_ucoeff = [(-32*sqrt(21)+180)/360, 0.0, 64*sqrt(21)/360, 0.0, (-32*sqrt(21) - 180)/360]
    ζ4_Fucoeff = [(9-sqrt(21))/360, 0.0, 64/360, 98/360, (9+sqrt(21))/360]

    x2 = collocation_constraint(x2_ucoeff, x2_Fucoeff, dt, ublock, Fublock)
    x4 = collocation_constraint(x4_ucoeff, x4_Fucoeff, dt, ublock, Fublock)
    ζ2 = collocation_constraint(ζ2_ucoeff, ζ2_Fucoeff, dt, ublock, Fublock)
    ζ4 = collocation_constraint(ζ4_ucoeff, ζ4_Fucoeff, dt, ublock, Fublock)

    vcat(x2, x4, ζ2, ζ4)
end

# function simpson_hermite_timestep_residual(vavf::VarAssimVectorField, ublock, ocblock, p, scblock, dt, λblock)
#     (;D) = vavf

#     R = 1
#     Fublock = [vavf.vf(ublock[r], ocblock[r], p, scblock[r]) for r=1:(R+2)] 

#     simp_ucoeff = [1.0, 0.0, -1.0]
#     simp_Fucoeff = [1.0/6, 4/6, 1/6]
#     herm_ucoeff = [0.5, -1.0, 0.5]
#     herm_Fucoeff = [1/8, 0.0, -1/8]

#     simp = collocation_constraint(simp_ucoeff, simp_Fucoeff, dt, ublock, Fublock)
#     herm = collocation_constraint(herm_ucoeff, herm_Fucoeff, dt, ublock, Fublock)

#     control_constraint = ocblock[2][1] - ((ocblock[1][1] + ocblock[3][1])/2.0 + (dt/8.0)*(ocblock[1][2] - ocblock[3][2]))

#     λblock .* vcat(simp, herm, control_constraint)
# end

# This function migth have a typo??
# function hlgl4_timestep_residual(vavf :: VarAssimVectorField, ublock, ocblock, p, scblock, dt, λblock)
#     (;D) = vavf

#     R = 3
#     Fublock = [vavf.vf(ublock[r], ocblock[r], p, scblock[r]) for r=1:(R+2)] 

#     x2_ucoeff = [(7sqrt(5)+9)/50, -1.0, 32/50, 0.0, (-7sqrt(5)+9)/50]
#     x2_Fucoeff = [(1+sqrt(5))/50, 0.0, 0.0, 0.0, (sqrt(5)-1)/50]
#     x4_ucoeff = [(-7sqrt(5)+9)/50, 0.0, 32/50, -1.0, (7sqrt(5)+9)/50]
#     x4_Fucoeff = [(1-sqrt(5))/50, 0.0, 0.0, 0.0, (-sqrt(5)-1)/50]

#     ζ2_ucoeff = [(32sqrt(5)+60)/120, 0.0, -72/120, 0.0, (32sqrt(5)-60)/120]
#     ζ2_Fucoeff = [5+3sqrt(5)/120, 50/120, 0.0, 0.0, (5-3sqrt(5))/120]
#     ζ4_ucoeff = [(-32sqrt(5)+60)/120, 0.0, 72/120, 0.0, (-32sqrt(5)-60)/120]
#     ζ4_Fucoeff = [(5-3sqrt(5))/120, 0.0, 0.0, 50/120, (5+3sqrt(5))/120]

#     x2 = collocation_constraint(x2_ucoeff, x2_Fucoeff, dt, ublock, Fublock)
#     x4 = collocation_constraint(x4_ucoeff, x4_Fucoeff, dt, ublock, Fublock)
#     ζ2 = collocation_constraint(ζ2_ucoeff, ζ2_Fucoeff, dt, ublock, Fublock)
#     ζ4 = collocation_constraint(ζ4_ucoeff, ζ4_Fucoeff, dt, ublock, Fublock)

#     control_constraint2 = ocblock[2][1] - ((ocblock[1][1] + ocblock[3][1])/2.0 + (dt/8.0)*(ocblock[1][2] - ocblock[3][2]))
#     control_constraint4 = ocblock[4][1] - ((ocblock[3][1] + ocblock[5][1])/2.0 + (dt/8.0)*(ocblock[3][2] - ocblock[5][2]))

#     λblock .* vcat(x2, x4, ζ2, ζ4, control_constraint2, control_constraint4) #length 4*D+2
# end


# function hlgl5_timestep_residual(vavf :: VarAssimVectorField, ublock, ocblock, p, scblock, dt, λblock)
#     (;D) = vavf

#     R = 3
#     Fublock = [vavf.vf(ublock[r], ocblock[r], p, scblock[r]) for r=1:(R+2)] 

#     x2_ucoeff = [(39*sqrt(21)+231)/686, -1.0, (224/686), 0.0, (-39*sqrt(21)+231)/686]
#     x2_Fucoeff = [(3*sqrt(21)+21)/686, 0.0, -16*sqrt(21)/686, 0.0, (3*sqrt(21) - 21)/686]
#     x4_ucoeff = [(-39*sqrt(21)+231)/686, 0.0, (224/686), -1.0, (39*sqrt(21)+231)/686]
#     x4_Fucoeff = [(-3*sqrt(21)+21)/686, 0.0, 16*sqrt(21)/686, 0.0, (-3*sqrt(21) - 21)/686]

#     ζ2_ucoeff = [(32*sqrt(21)+180)/360, 0.0, -64*sqrt(21)/360, 0.0, (32*sqrt(21) - 180)/360]
#     ζ2_Fucoeff = [(9+sqrt(21))/360, 98/360, 64/360, 0.0, (9-sqrt(21))/360]
#     ζ4_ucoeff = [(-32*sqrt(21)+180)/360, 0.0, 64*sqrt(21)/360, 0.0, (-32*sqrt(21) - 180)/360]
#     ζ4_Fucoeff = [(9-sqrt(21))/360, 0.0, 64/360, 98/360, (9+sqrt(21))/360]

#     x2 = collocation_constraint(x2_ucoeff, x2_Fucoeff, dt, ublock, Fublock)
#     x4 = collocation_constraint(x4_ucoeff, x4_Fucoeff, dt, ublock, Fublock)
#     ζ2 = collocation_constraint(ζ2_ucoeff, ζ2_Fucoeff, dt, ublock, Fublock)
#     ζ4 = collocation_constraint(ζ4_ucoeff, ζ4_Fucoeff, dt, ublock, Fublock)

#     control_constraint2 = ocblock[2][1] - ((ocblock[1][1] + ocblock[3][1])/2.0 + (dt/8.0)*(ocblock[1][2] - ocblock[3][2]))
#     control_constraint4 = ocblock[4][1] - ((ocblock[3][1] + ocblock[5][1])/2.0 + (dt/8.0)*(ocblock[3][2] - ocblock[5][2]))

#     λblock .* vcat(x2, x4, ζ2, ζ4, control_constraint2, control_constraint4) #length 4*D+2
# end

